from itertools import cycle
from typing import List

import numpy as np
import pandas as pd
import streamlit as st
from faker import Faker
from faker.providers import BaseProvider

from .common import st_command_with_default

_DATAFRAME_CATEGORIES_DICT = {
    "animals": [
        "cow",
        "rabbit",
        "duck",
        "shrimp",
        "pig",
        "goat",
        "crab",
        "deer",
        "bee",
        "sheep",
        "fish",
        "turkey",
        "dove",
        "chicken",
        "horse",
    ],
    "names": [
        "James",
        "Mary",
        "Robert",
        "Patricia",
        "John",
        "Jennifer",
        "Michael",
        "Linda",
        "William",
        "Elizabeth",
        "Ahmed",
        "Barbara",
        "Richard",
        "Susan",
        "Salomon",
        "Juan Luis",
    ],
    "cities": [
        "Stockholm",
        "Denver",
        "Moscow",
        "Marseille",
        "Palermo",
        "Tokyo",
        "Lisbon",
        "Oslo",
        "Nairobi",
        "Río de Janeiro",
        "Berlin",
        "Bogotá",
        "Manila",
        "Madrid",
        "Milwaukee",
    ],
    "colors": [
        "red",
        "orange",
        "yellow",
        "green",
        "blue",
        "indigo",
        "purple",
        "pink",
        "silver",
        "gold",
        "beige",
        "brown",
        "grey",
        "black",
        "white",
    ],
}

fake = Faker()


class StreamlitDataDisplayProvider(BaseProvider):
    def _metric_unit(self) -> str:
        return self.random_element(
            ["s", "ms", "%", "$", "€", "USD", "users", "°C", "°F"]
        )

    def _sign(self) -> str:
        return self.random_element(["+", "-"])

    @staticmethod
    def _pandas_dataframe_builder(
        size: int, cols: List, col_names: List = None, intervals=None, seed=None
    ):

        default_intervals = {
            "i": (0, 10),
            "f": (0, 100),
            "c": ("names", 5),
            "d": ("2020-01-01", "2020-12-31"),
        }

        rng = np.random.default_rng(seed)

        first_c = default_intervals["c"][0]
        categories_names = cycle(
            [first_c] + [c for c in _DATAFRAME_CATEGORIES_DICT.keys() if c != first_c]
        )

        default_intervals["c"] = (
            categories_names,
            default_intervals["c"][1],
        )

        if isinstance(col_names, list):
            assert len(col_names) == len(cols), (
                f"The fake DataFrame should have {len(cols)} columns but col_names"
                f" is a list with {len(col_names)} elements"
            )
        elif col_names is None:
            suffix = {"c": "cat", "i": "int", "f": "float", "d": "date"}
            col_names = [
                f"column_{str(i)}_{suffix.get(col)}" for i, col in enumerate(cols)
            ]

        if isinstance(intervals, list):
            assert len(intervals) == len(cols), (
                f"The fake DataFrame should have {len(cols)} columns but intervals"
                f" is a list with {len(intervals)} elements"
            )
        else:
            if isinstance(intervals, dict):
                assert (
                    len(set(intervals.keys()) - set(default_intervals.keys())) == 0
                ), "The intervals parameter has invalid keys"
                default_intervals.update(intervals)
            intervals = [default_intervals[col] for col in cols]
        df = pd.DataFrame()
        for col, col_name, interval in zip(cols, col_names, intervals):
            if interval is None:
                interval = default_intervals[col]
            assert (len(interval) == 2 and isinstance(interval, tuple)) or isinstance(
                interval, list
            ), (
                f"This interval {interval} is neither a tuple of two elements nor"
                " a list of strings."
            )
            if col in ("i", "f", "d"):
                start, end = interval
            if col == "i":
                df[col_name] = rng.integers(start, end, size)
            elif col == "f":
                df[col_name] = rng.uniform(start, end, size)
            elif col == "c":
                if isinstance(interval, list):
                    categories = np.array(interval)
                else:
                    cat_family, length = interval
                    if isinstance(cat_family, cycle):
                        cat_family = next(cat_family)
                    assert cat_family in _DATAFRAME_CATEGORIES_DICT.keys(), (
                        f"There are no samples for category '{cat_family}'."
                        " Consider passing a list of samples or use one of the"
                        f" available categories: {_DATAFRAME_CATEGORIES_DICT.keys()}"
                    )
                    categories = rng.choice(
                        _DATAFRAME_CATEGORIES_DICT[cat_family],
                        length,
                        replace=False,
                        shuffle=True,
                    )
                df[col_name] = rng.choice(categories, size, shuffle=True)
            elif col == "d":
                df[col_name] = rng.choice(pd.date_range(start, end), size)
        return df

    def metric(self, **kwargs):
        unit = self._metric_unit()
        return st_command_with_default(
            st.metric,
            {
                "label": fake.word().title(),
                "value": f"{fake.random_int()} {unit}",
                "delta": f"{self._sign()} {fake.random_int(1, 100)} {unit}",
            },
            **kwargs,
        )

    def json(self, **kwargs):
        return st_command_with_default(
            st.json,
            {
                "body": fake.simple_profile(),
                "expanded": True,
            },
            **kwargs,
        )

    def _pandas_dataframe(self):
        size = self.random_int(5, 20)
        num_cols = self.random_int(3, 6)
        col_names = [fake.name() for _ in range(num_cols)]
        cols = [self.random_element(list("cifd")) for _ in range(num_cols)]
        return self._pandas_dataframe_builder(
            size=size,
            cols=cols,
            col_names=col_names,
        )

    def dataframe(self, **kwargs):
        return st_command_with_default(
            st.dataframe,
            {"data": self._pandas_dataframe()},
            **kwargs,
        )

    def table(self, **kwargs):
        return st_command_with_default(
            st.table,
            {"data": self._pandas_dataframe()},
            **kwargs,
        )



import altair as alt
import numpy as np
import pandas as pd
import streamlit as st
from faker.providers import BaseProvider
from matplotlib import pyplot as plt
from streamlit_extras import altex

from .common import st_command_with_default


@st.experimental_memo
def url_to_dataframe(url: str, parse_dates: list = ["date"]) -> pd.DataFrame:
    """Collects a CSV/JSON file from a URL and load it into a dataframe, with appropriate caching (memo)
    Args:
        url (str): URL of the CSV/JSON file
        parse_dates (list): Columns where date parsing should be done
    Returns:
        pd.DataFrame: Resulting dataframe
    """
    if url.endswith(".csv"):
        return pd.read_csv(url, parse_dates=parse_dates)
    elif url.endswith(".json"):
        return pd.read_json(url)
    else:
        raise Exception("URL must end with .json or .csv")


SEATTLE_WEATHER_URL = (
    "https://raw.githubusercontent.com/tvst/plost/master/data/seattle-weather.csv"
)
SP500_URL = "https://raw.githubusercontent.com/tvst/plost/master/data/sp500.csv"

STOCKS_DATA_URL = (
    "https://raw.githubusercontent.com/vega/vega/main/docs/data/stocks.csv"
)
BARLEY_DATA_URL = (
    "https://raw.githubusercontent.com/vega/vega/main/docs/data/barley.json"
)


@st.experimental_singleton()
def get_datasets():
    N = 50
    rand = pd.DataFrame()
    rand["a"] = np.arange(N)
    rand["b"] = np.random.rand(N)
    rand["c"] = np.random.rand(N)

    N = 500
    events = pd.DataFrame()
    events["time_delta_s"] = np.random.randn(N)
    events["servers"] = np.random.choice(["server 1", "server 2", "server 3"], N)

    N = 500
    randn = pd.DataFrame(
        np.random.randn(N, 4),
        columns=["a", "b", "c", "d"],
    )

    N = 200
    pageviews = pd.DataFrame()
    pageviews["pagenum"] = [f"page-{i:03d}" for i in range(N)]
    pageviews["pageviews"] = np.random.randint(0, 1000, N)

    return dict(
        rand=rand,
        randn=randn,
        events=events,
        pageviews=pageviews,
        stocks=url_to_dataframe(STOCKS_DATA_URL),
        seattle_weather=url_to_dataframe(SEATTLE_WEATHER_URL),
        sp500=url_to_dataframe(SP500_URL),
        barley=url_to_dataframe(BARLEY_DATA_URL),
    )


datasets = get_datasets()


@st.experimental_memo
def _line_chart():
    st.line_chart(
        data=datasets["stocks"].query("symbol == 'GOOG'"),
        x="date",
        y="price",
        # title="A beautiful simple line chart",
    )


@st.experimental_memo
def _multi_line_chart():
    altex.line_chart(
        data=datasets["stocks"],
        x="date",
        y="price",
        color="symbol",
        # title="A beautiful multi line chart",
    )


def _bar_chart():
    symbol = np.random.choice(datasets["stocks"].symbol.unique(), 1)[0]
    st.bar_chart(
        data=datasets["stocks"].query(
            f"(symbol == '{symbol}') & (date >= '2006-01-01')"
        ),
        x="date",
        y="price",
        # title="A beautiful bar chart",
    )


@st.experimental_memo
def _hist_chart():
    altex.hist_chart(
        data=datasets["stocks"].assign(price=datasets["stocks"].price.round(0)),
        x="price",
        title="A beautiful histogram",
    )


@st.experimental_memo
def _scatter_chart():
    altex.scatter_chart(
        data=datasets["seattle_weather"],
        x=alt.X("wind:Q", title="Custom X title"),
        y=alt.Y("temp_min:Q", title="Custom Y title"),
        # title="A beautiful scatter chart with custom opacity",
        opacity=0.2,
    )


@st.experimental_memo
def _bar_chart_horizontal():
    altex.bar_chart(
        data=datasets["seattle_weather"].head(15),
        x="temp_max:Q",
        y=alt.Y("date:O", title="Temperature"),
        # title="A beautiful horizontal bar chart",
    )


@st.experimental_memo
def _bar_chart_log():
    altex.bar_chart(
        data=datasets["seattle_weather"],
        x=alt.X("temp_max:Q", title="Temperature"),
        y=alt.Y(
            "count()",
            title="Count of records",
            scale=alt.Scale(type="symlog"),
        ),
        # title="A beautiful histogram... with log scale",
    )


@st.experimental_memo
def _bar_chart_sorted():
    altex.bar_chart(
        data=datasets["seattle_weather"]
        .sort_values(by="temp_max", ascending=False)
        .head(25),
        x=alt.X("date", sort="-y"),
        y=alt.Y("temp_max:Q"),
        # title="A beautiful sorted-by-value bar chart",
    )


@st.experimental_memo
def _time_heatmap_chart():
    altex.hist_chart(
        data=datasets["seattle_weather"],
        x="week(date):T",
        y="day(date):T",
        color=alt.Color(
            "median(temp_max):Q",
            legend=None,
        ),
        # title="A beautiful time hist chart",
    )


@st.experimental_memo
def _sparkline_chart():
    altex.line_chart(
        data=datasets["stocks"].query("symbol == 'GOOG'"),
        x="date",
        y="price",
        # title="A beautiful sparkline chart",
        rolling=7,
        height=150,
    )


@st.experimental_memo
def _sparkbar_chart():
    altex.bar_chart(
        data=datasets["stocks"].query("symbol == 'GOOG'"),
        x="date",
        y="price",
        # title="A beautiful sparkbar chart",
        height=150,
    )


@st.experimental_memo
def _bar_stacked_chart():
    altex.bar_chart(
        data=datasets["barley"],
        x=alt.X("variety", title="Variety"),
        y="sum(yield)",
        color="site",
        # title="A beautiful stacked bar chart",
    )


@st.experimental_memo
def _bar_normalized_chart():
    altex.bar_chart(
        data=datasets["barley"],
        x=alt.X("variety:N", title="Variety"),
        y=alt.Y("sum(yield):Q", stack="normalize"),
        color="site:N",
        # title="A beautiful normalized stacked bar chart",
    )


@st.experimental_memo
def _bar_grouped_chart():
    altex.bar_chart(
        data=datasets["barley"],
        x="year:O",
        y="sum(yield):Q",
        color="year:N",
        column="site:N",
        # title="A beautiful grouped bar charts",
        width=90,
        use_container_width=False,
    )


class StreamlitChartProvider(BaseProvider):
    def altair_chart(self):
        return self.random_element(
            [
                _bar_grouped_chart,
                _bar_chart_horizontal,
                _bar_chart_log,
                _bar_chart_sorted,
                _bar_stacked_chart,
                _scatter_chart,
                _sparkline_chart,
                _sparkbar_chart,
                _time_heatmap_chart,
                _hist_chart,
                _bar_normalized_chart,
            ]
        )()

    def line_chart(self):
        return self.random_element([_line_chart, _multi_line_chart])()

    def bar_chart(self):
        return self.random_element([_bar_chart])()

    def map(self, **kwargs):
        return st_command_with_default(
            st.map,
            {
                "data": pd.DataFrame(
                    np.random.randn(self.random_int(600, 1000), 2) / [50, 50]
                    + [37.76, -122.4],
                    columns=["lat", "lon"],
                )
            },
            **kwargs,
        )

    def pyplot(self, **kwargs):
        arr = np.random.normal(1, 1, size=100)
        fig, ax = plt.subplots()
        ax.hist(arr, bins=20)
        return st_command_with_default(st.pyplot, {"fig": fig}, **kwargs)

    def vega_lite_chart(self, **kwargs):
        data = pd.DataFrame(np.random.randn(200, 3), columns=["a", "b", "c"])
        return st_command_with_default(
            st.vega_lite_chart,
            {
                "data": data,
                "spec": {
                    "mark": {"type": "circle", "tooltip": True},
                    "encoding": {
                        "x": {"field": "a", "type": "quantitative"},
                        "y": {"field": "b", "type": "quantitative"},
                        "size": {"field": "c", "type": "quantitative"},
                        "color": {"field": "c", "type": "quantitative"},
                    },
                },
            },
            **kwargs,
        )

    def plotly_chart(self):
        raise NotImplementedError

    def bokeh_chart(self):
        raise NotImplementedError

    def pydeck_chart(self):
        raise NotImplementedError

    def graphviz_chart(self):
        raise NotImplementedError



import streamlit as st
from faker import Faker
from faker.providers import BaseProvider, date_time

from .common import st_command_with_default

fake = Faker()
fake.add_provider(date_time)


class StreamlitMediaProvider(BaseProvider):
    def image(self, **kwargs):
        return st_command_with_default(
            st.image,
            {
                "image": f"http://placekitten.com/{self.random_int(118, 120)}/{self.random_int(118, 120)}"
            },
            **kwargs,
        )

    def video(self, **kwargs):
        raise NotImplementedError

    def audio(self, **kwargs):
        raise NotImplementedError



import streamlit as st
from faker import Faker
from faker.providers import BaseProvider, date_time

from .common import st_command_with_default

fake = Faker()
fake.add_provider(date_time)


class StreamlitInputProvider(BaseProvider):
    def text_input(self, **kwargs):
        return st_command_with_default(
            st.text_input,
            {
                "label": fake.word().title(),
                "value": fake.sentence(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def text_area(self, **kwargs):
        return st_command_with_default(
            st.text_area,
            {
                "label": fake.word().title(),
                "value": fake.text(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def button(self, **kwargs):
        return st_command_with_default(
            st.button,
            {
                "label": fake.name(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def download_button(self, **kwargs):
        return st_command_with_default(
            st.download_button,
            {
                "label": "Download " + fake.name(),
                "data": "foo",
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def checkbox(self, **kwargs):
        return st_command_with_default(
            st.checkbox,
            {
                "label": fake.sentence(),
                "value": self.random_element((True, False)),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def radio(self, **kwargs):
        return st_command_with_default(
            st.radio,
            {
                "label": fake.name(),
                "options": fake.sentence().lower().split(),
                "index": 0,
                "key": fake.name().lower(),
            },
        )

    def selectbox(self, **kwargs):
        return st_command_with_default(
            st.selectbox,
            {
                "label": fake.name(),
                "options": fake.sentence().lower().split(),
                "index": 0,
                "key": fake.name().lower(),
            },
        )

    def multiselect(self, **kwargs):
        options = fake.sentence().lower().split()
        return st_command_with_default(
            st.multiselect,
            {
                "label": fake.name(),
                "options": options,
                "default": options[0],
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def slider(self, **kwargs):
        return st_command_with_default(
            st.slider,
            {
                "label": fake.name(),
                "min_value": self.random_int(0, 10),
                "max_value": self.random_int(50, 100),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def select_slider(self, **kwargs):
        return st_command_with_default(
            st.select_slider,
            {
                "label": fake.name(),
                "options": self.random_elements(
                    [fake.name() for _ in range(self.random_int(5, 10))],
                    self.random_int(3, 6),
                ),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def number_input(self, **kwargs):
        return st_command_with_default(
            st.number_input,
            {
                "label": fake.name(),
                "min_value": self.random_int(0, 10),
                "max_value": self.random_int(50, 100),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def date_input(self, **kwargs):
        return st_command_with_default(
            st.date_input,
            {
                "label": fake.name(),
                "value": fake.date_this_year(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def time_input(self, **kwargs):
        return st_command_with_default(
            st.time_input,
            {
                "label": fake.name(),
                "value": fake.date_time(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def file_uploader(self, **kwargs):
        return st_command_with_default(
            st.file_uploader,
            {
                "label": fake.name(),
                "type": self.random_element((".png", ".csv", ".txt", ".parquet")),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def camera_input(self, **kwargs):
        return st_command_with_default(
            st.camera_input,
            {
                "label": fake.name(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )

    def color_picker(self, **kwargs):
        return st_command_with_default(
            st.color_picker,
            {
                "label": fake.name(),
                "key": fake.name().lower(),
            },
            **kwargs,
        )



from functools import partial
from typing import Callable


def st_command_with_default(st_command: Callable, kwarg_defaults: dict, **kwargs):
    for kwarg, default in kwarg_defaults.items():
        kwargs[kwarg] = kwargs.get(kwarg, default)
    return partial(st_command, **kwargs)()



from faker import Faker

from .chart import StreamlitChartProvider
from .data_display import StreamlitDataDisplayProvider
from .input import StreamlitInputProvider
from .media import StreamlitMediaProvider
from .status import StreamlitStatusProvider
from .text import StreamlitTextProvider


def get_streamlit_faker(seed: int = None, locale: str = "en-US"):
    fake = Faker(locale=locale)
    fake.add_provider(StreamlitTextProvider)
    fake.add_provider(StreamlitChartProvider)
    fake.add_provider(StreamlitInputProvider)
    fake.add_provider(StreamlitDataDisplayProvider)
    fake.add_provider(StreamlitStatusProvider)
    fake.add_provider(StreamlitMediaProvider)

    _seed = seed if seed else fake.random_int(0, 1_000)
    Faker.seed(_seed)

    return fake


StreamlitFaker = get_streamlit_faker()

text_commands = list(set(dir(StreamlitTextProvider)) - set(dir(StreamlitChartProvider)))
chart_commands = list(
    set(dir(StreamlitChartProvider)) - set(dir(StreamlitTextProvider))
)
data_display_commands = list(
    set(dir(StreamlitDataDisplayProvider)) - set(dir(StreamlitTextProvider))
)
input_commands = list(
    set(dir(StreamlitInputProvider)) - set(dir(StreamlitTextProvider))
)
status_commands = list(
    set(dir(StreamlitStatusProvider)) - set(dir(StreamlitTextProvider))
)
media_commands = list(
    set(dir(StreamlitMediaProvider)) - set(dir(StreamlitTextProvider))
)

all_commands = (
    text_commands
    + chart_commands
    + input_commands
    + status_commands
    + data_display_commands
    + media_commands
)



from multiprocessing import AuthenticationError

import streamlit as st
from faker import Faker
from faker.providers import BaseProvider

from .common import st_command_with_default

fake = Faker()


class StreamlitStatusProvider(BaseProvider):
    @staticmethod
    def error(**kwargs):
        return st_command_with_default(st.error, {"body": fake.sentence()}, **kwargs)

    @staticmethod
    def success(**kwargs):
        return st_command_with_default(st.success, {"body": fake.sentence()}, **kwargs)

    @staticmethod
    def warning(**kwargs):
        return st_command_with_default(st.warning, {"body": fake.sentence()}, **kwargs)

    @staticmethod
    def info(**kwargs):
        return st_command_with_default(st.info, {"body": fake.sentence()}, **kwargs)

    def exception(self, **kwargs):
        return st_command_with_default(
            st.exception,
            {
                "exception": self.random_element(
                    [AttributeError, KeyError, AuthenticationError, ImportError]
                )
            },
            **kwargs
        )

    @staticmethod
    def balloons():
        return st.balloons()

    @staticmethod
    def snow():
        return st.snow()



import streamlit as st
from faker import Faker
from faker.providers import BaseProvider

from .common import st_command_with_default

fake = Faker()

DEFAULT_CODE = """
def jpeg_res(filename):
    # open image for reading in binary mode
    with open(filename,'rb') as img_file:

        # height of image (in 2 bytes) is at 164th position
        img_file.seek(163)

        # read the 2 bytes
        a = img_file.read(2)

        # calculate height
        height = (a[0] << 8) + a[1]
"""

DEFAULT_LATEX = r"""a + ar + a r^2 + a r^3 + \cdots + a r^{n-1} =
\sum_{k=0}^{n-1} ar^k =
a \left(\frac{1-r^{n}}{1-r}\right)"""


class StreamlitTextProvider(BaseProvider):
    def markdown(self, **kwargs):
        return st_command_with_default(st.markdown, {"body": fake.text()}, **kwargs)

    def text(self, **kwargs):
        return st_command_with_default(st.text, {"body": fake.text()}, **kwargs)

    def caption(self, **kwargs):
        return st_command_with_default(st.caption, {"body": fake.sentence()}, **kwargs)

    def title(self, **kwargs):
        return st_command_with_default(
            st.title, {"body": fake.sentence().title()}, **kwargs
        )

    def header(self, **kwargs):
        return st_command_with_default(
            st.header, {"body": fake.sentence().title()}, **kwargs
        )

    def subheader(self, **kwargs):
        return st_command_with_default(
            st.subheader, {"body": fake.sentence().title()}, **kwargs
        )

    def code(self, **kwargs):
        return st_command_with_default(st.code, {"body": DEFAULT_CODE}, **kwargs)

    def latex(self, **kwargs):
        return st_command_with_default(st.latex, {"body": DEFAULT_LATEX}, **kwargs)
